#pragma once

void mixChordSample(void);
void recalcChordLength(void);
void resetChord(void);
void undoChord(void);
void toggleChordLength(void);
void setChordMajor(void);
void setChordMinor(void);
void setChordSus4(void);
void setChordMajor6(void);
void setChordMinor6(void);
void setChordMajor7(void);
void setChordMinor7(void);
void selectChordNote1(void);
void selectChordNote2(void);
void selectChordNote3(void);
void selectChordNote4(void);
